/* This sets bit 4 in the flags of the 'head' table of a TrueType font.

   http://developer.apple.com/fonts/TTRefMan/ - Apple's TrueType specification.
   http://www.microsoft.com/typography/SpecificationsOverview.mspx - Microsoft's TrueType specification.
   The two specifications unfortunately differ in a number of ways, but this software is unaffected. */

/* This has been written with clarity and portability in mind, rather than performance. In particular:
   - The C language is used here, instead of something higher-level or more powerful. stdint.h comes from ISO/IEC 9899:1999 
     ("C99"), the syntax and other headers conform to ISO/IEC 9899:1990 ("C89").
   - make_uint32, big_endian_*, and unaligned_read/write_* on some platforms could have much faster/smaller implementations. In
     many cases these can be written as a single assembler assembly language instruction.
   - Memory-mapped files are often faster than conventional read()/write() I/O when available. 
   - Native low-level I/O (open(2), CreateFile) tends to be faster than fopen(). */

/* Upon success, no output is generated. */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef _MSC_VER

typedef unsigned __int8 uint8_t;
typedef unsigned __int16 uint16_t;
typedef unsigned __int32 uint32_t;
typedef unsigned __int64 uint64_t;

typedef __int16 int16_t;
typedef __int32 int32_t;
typedef __int64 int64_t;

#else

/* This is a standard C99 header, but not all implementations of the C standard library are C99-compatible. If yours is not, 
   simply add typedefs in accordance with the C99 definitions. See http://en.wikipedia.org/wiki/Stdint.h for details. */
#	include <stdint.h>

#endif

static uint32_t make_uint32(uint8_t b0, uint8_t b1, uint8_t b2, uint8_t b3)
{
	union
	{
		uint8_t bytes[4];
		uint32_t uint;
	} result;

	result.bytes[0] = b0;
	result.bytes[1] = b1;
	result.bytes[2] = b2;
	result.bytes[3] = b3;

	return result.uint;
}

#ifdef _MSC_VER
#	pragma warning(push)
#	pragma warning(disable: 4701) // potentially uninitialized local variable 'result' used
#endif

#define DEFINE_BIG_ENDIAN_FUNC(bits) \
	static uint##bits##_t big_endian_##bits(uint##bits##_t x) \
	{ \
		typedef uint##bits##_t uint_t; \
		unsigned size = sizeof(uint_t) / sizeof(uint8_t); \
		union \
		{ \
			uint8_t bytes[sizeof(uint_t) / sizeof(uint8_t)]; \
			uint_t uint; \
		} result; \
		\
		unsigned i = 0; \
		for(; i != size; ++i) \
			result.bytes[i] = (uint8_t)(x >> (8 * (size - 1 - i))); \
		\
		return result.uint; \
	}

DEFINE_BIG_ENDIAN_FUNC(16)
DEFINE_BIG_ENDIAN_FUNC(32)
DEFINE_BIG_ENDIAN_FUNC(64)

#define DEFINE_UNALIGNED_READ_FUNC(bits) \
	static uint##bits##_t unaligned_read_##bits(const uint##bits##_t *value) \
	{ \
		typedef uint##bits##_t uint_t; \
		union \
		{ \
			unsigned char bytes[sizeof(uint_t)]; \
			uint_t uint; \
		} result; \
		const unsigned char *src = (const unsigned char *)value; \
		unsigned i = 0; \
		for(; i != sizeof(uint_t); ++i) \
			result.bytes[i] = src[i]; \
		return result.uint; \
	}

DEFINE_UNALIGNED_READ_FUNC(16)
DEFINE_UNALIGNED_READ_FUNC(32)

#ifdef _MSC_VER
#	pragma warning(pop) // potentially uninitialized local variable 'result' used
#endif

#define DEFINE_UNALIGNED_WRITE_FUNC(bits) \
	static void unaligned_write_##bits(uint##bits##_t *dest, uint##bits##_t src) \
	{ \
		typedef uint##bits##_t uint_t; \
		union \
		{ \
			uint_t uint; \
			unsigned char bytes[sizeof(uint_t)]; \
		} src_union; \
		\
		unsigned char *dest_bytes = (unsigned char *)dest; \
		unsigned i = 0; \
		src_union.uint = src; \
		for(; i != sizeof(uint_t); ++i) \
			dest_bytes[i] = src_union.bytes[i]; \
	}

DEFINE_UNALIGNED_WRITE_FUNC(16)
DEFINE_UNALIGNED_WRITE_FUNC(32)
DEFINE_UNALIGNED_WRITE_FUNC(64)

/* Typedefs for TrueType file structures. */

#if defined __GNUC__
#	define ATTRIBUTE_PACK __attribute__((packed))
#elif defined _MSC_VER || defined __WATCOMC__
#	define ATTRIBUTE_PACK
#	pragma pack(push, 1)
#else
#	error Unsupported compiler.
#endif

typedef int16_t FWord;
typedef int32_t Fixed;
typedef int64_t longDateTime;

struct offset_subtable
{
	uint32_t scaler_type;
	uint16_t numTables;
	uint16_t searchRange;
	uint16_t entrySelector;
	uint16_t rangeShift;
} ATTRIBUTE_PACK;

struct table_directory
{
	uint32_t tag;
	uint32_t checkSum;
	uint32_t offset;
	uint32_t length;
} ATTRIBUTE_PACK;

/* The _1 suffix refers to version 1. */
struct head_table_1
{
	Fixed version;
	Fixed fontRevision;
	uint32_t checkSumAdjustment;
	uint32_t magicNumber;
	uint16_t flags;
	uint16_t unitsPerEm;
	longDateTime created;
	longDateTime modified;
	FWord xMin;
	FWord yMin;
	FWord xMax;
	FWord yMax;
	uint16_t macStyle;
	uint16_t lowestRecPPEM;
	int16_t fontDirectionHint;
	int16_t indexToLocFormat;
	int16_t glyphDataFormat;
} ATTRIBUTE_PACK;

static uint32_t calc_table_checksum(uint32_t *table, uint32_t numberOfBytesInTable)
{
	uint32_t sum = 0;
	uint32_t nLongs = (numberOfBytesInTable + 3) / 4;

	while(nLongs-- > 0)
	{
		sum += big_endian_32(unaligned_read_32(table));
		++table;
	}

	return sum;
}

int main(int argc, char **argv)
{
	int result = EXIT_FAILURE;

	assert(sizeof(uint32_t) == sizeof(uint8_t) * 4);
	assert(sizeof(uint16_t) == sizeof(uint8_t) * 2);

	assert(sizeof(struct offset_subtable) == sizeof(uint8_t) * 12);
	assert(sizeof(struct table_directory) == sizeof(uint8_t) * 16);
	assert(sizeof(struct head_table_1) == sizeof(uint8_t) * 54);

	if(argc != 2)
	{
		fputs("Usage: ", stdout);
		fputs(argv[0], stdout);
		fputs(" font_name\n", stdout);
		return EXIT_FAILURE;
	}

	{
		FILE *ttf_file = fopen(argv[1], "r+b");

		if(!ttf_file)
		{
			perror("Unable to open font");
			return EXIT_FAILURE;
		}

		/* A somewhat convoluted way to get the size of the file starts here.
		   */
		if(fseek(ttf_file, 0, SEEK_END))
		{
			perror("Error seeking to end of file");
		}
		else
		{
			/* Many platforms are capable of handling files larger than what a 
			   long can hold (especially in the case of a 32-bit platform with 
			   32-bit longs reading files greater than 2 GB), and offer 
			   modified versions of ftell() that support these larger files.
			   
			   These versions of ftell() will not be used here. If ftell() 
			   needs to return a value larger than 2 GB, something has 
			   obviously gone terribly wrong. The check for the error result 
			   of -1L should catch this and abort the program, if it ever 
			   happens. */

			long ttf_size = ftell(ttf_file);
			assert(ttf_size >= -1L);
			if(ttf_size == -1L)
			{
				perror("Error getting the font size");
			}
			else if((unsigned long)ttf_size < sizeof(struct offset_subtable))
			{
				fputs("Specified font is not a valid TrueType font.\n", stderr);
			}
			else
			{
				/* The checksum requires that the font data be 32-bit padded and filled with zeros. */
				struct offset_subtable *header = malloc(ttf_size + 3);
				if(!header)
				{
					fputs("Not enough memory.\n", stderr);
				}
				else
				{
					rewind(ttf_file);
					
					/* All memory accesses are aligned for the offset_subtable and table_directory. */

					if(!fread(header, ttf_size, 1, ttf_file))
					{
						fputs("Font is truncated.\n", stderr);
					}
					else if(header->scaler_type != make_uint32('t', 'r', 'u', 'e') && 
						header->scaler_type != make_uint32(0, 1, 0, 0))
					{
						fputs("Font is not a valid TrueType font.\n", stderr);
					}
					else 
					{
						uint16_t num_tables = big_endian_16(header->numTables);
						if((unsigned long)ttf_size < sizeof(struct offset_subtable) + num_tables * sizeof(struct table_directory))
						{
							fputs("Font is truncated.\n", stderr);
						}
						else
						{
							/* There's a few things that could be validated here:
							   - Overlapping tables
							   - Invalid checksums 
							   - Duplicate tables */
							struct table_directory *tables = (struct table_directory *)(header + 1);
							uint16_t t = 0;
							for(;;)
							{
								if(t == num_tables)
								{
									fputs("Font has no 'head' table.\n", stderr);
									break;
								}

								if(tables[t].tag == make_uint32('h', 'e', 'a', 'd'))
								{
									uint32_t offset = big_endian_32(tables[t].offset), length = big_endian_32(tables[t].length);

									if((unsigned long)ttf_size < (offset + length) * sizeof(uint8_t))
									{
										fputs("Font is truncated.\n", stderr);
										break;
									}
									
									if(length < sizeof(Fixed))
									{
										fputs("'head' table is invalid.\n", stderr);
										break;
									}

									{
										struct head_table_1 *head = (struct head_table_1 *)((uint8_t *)header + offset);

										if(unaligned_read_32((const uint32_t *)&head->version) != (Fixed)make_uint32(0, 1, 0, 0))
										{
											fputs("'head' table is not version 1.0.\n", stderr);
											break;
										}

										if(length != sizeof(struct head_table_1))
										{
											fputs("'head' table is invalid.\n", stderr);
											break;
										}

										if(unaligned_read_32(&head->magicNumber) != make_uint32(0x5F, 0x0F, 0x3C, 0xF5))
										{
											fputs("Magic number in 'head' table is incorrect.", stderr);
											break;
										}

										/* Bit 4 - instructions may alter advance width (the advance widths might not scale linearly) */
										unaligned_write_16(&head->flags, unaligned_read_16(&head->flags) | big_endian_16(1 << 4));
										unaligned_write_64((uint64_t *)&head->modified, big_endian_64((uint64_t)time(NULL) + 24107ul * 24ul * 60ul * 60ul));

										unaligned_write_32(&head->checkSumAdjustment, 0);
										tables[t].checkSum = big_endian_32(calc_table_checksum((uint32_t *)head, length));
										((uint8_t *)header)[ttf_size + 0] = 0;
										((uint8_t *)header)[ttf_size + 1] = 0;
										((uint8_t *)header)[ttf_size + 2] = 0;
										unaligned_write_32(&head->checkSumAdjustment, big_endian_32(0xB1B0AFBA - calc_table_checksum((uint32_t *)header, ttf_size)));

										rewind(ttf_file);
										if(!fwrite(header, ttf_size, 1, ttf_file))
											fputs("Error writing font.\n", stderr);
									}

									break;
								}

								++t;
							}
						}
					}

					free(header);
				}
			}
		}

		fclose(ttf_file);
	}

	return result;
}
